open Ast
open Sast

exception Error of string;;

type environment = {
	functions: func_decl list;
	scope: string;
	locals: var_scope;
	globals: var_scope;
	has_return: bool;
	return_val: expr;
	return_type: var_type;
}

and var_scope = {
	prims: (string * var_type * expr) list;
	lists: (string * var_type * expr list) list
}

let beginning_scope = { prims = []; lists = [] }

let beginning_environment = { functions = []; globals = beginning_scope; locals = beginning_scope; scope = "global"; has_return = false; return_val = Id("None"); return_type = None}

let rec print_type ty =
	match ty with
		String -> "string"
		| Int -> "int"
		| Double -> "double"
		| Bool -> "bool"
		| Li(t) -> print_type t ^ " list"
		| None -> "none"

let get_id_from_expr ex = 
	match ex with
		Id(v) -> v
		| _ -> raise(Error("Trying to get id from a non-id expression\n"))


let check_arithmetic_binary_op t1 t2 =
	match (t1, t2) with
	| (Int, Int) -> Int
	| (Int, Double) -> Double
	| (Double, Int) -> Double
	| (Double, Double) -> Double
	| (_,_) -> raise(Error("Binary operation fails, wrong element type"))

let check_equality t1 t2 =
	prerr_string(print_type t1 ^ " " ^print_type t2 ^"\n");
	if t1 = t2 then Bool else
	match (t1, t2) with
	| (Int, Double) -> Bool
	| (Double, Int) -> Bool
	| (_, _) -> raise(Error("Equality operation fails, arguments not same type"))

let check_logic t1 t2 =
	match(t1, t2) with
	| (Int, Int) -> Bool
	| (Int, Double) -> Bool
	| (Double, Int) -> Bool
	| (Double, Double) -> Bool
	| (String, String) -> Bool
	| (_,_) -> raise(Error("Logical operation fails, arguments not of correct types"))

let get_literal_type l = match l with
	Int_Literal(i) -> Int
	| Double_Literal(d) -> Double
	| String_Literal(s) -> String
	| Bool_Literal(b) -> Bool
	| _ -> raise(Error("Should not be using _ here"))

let get_type_from_id var_table id =
	 if List.exists (fun (lid,_,_) -> lid=id) var_table.lists then
	 (let (_,ty,_) = List.find(fun (lid,_,_) -> lid=id) var_table.lists in ty)
	else
		let (_,ty,_) = try List.find (fun (vid, _, _) -> vid=id) var_table.prims with
		Not_found -> raise Not_found in ty

let check_for_var_existence var_table id =
	(List.exists(fun (vid,_,_) -> vid=id) var_table.prims
	|| List.exists(fun (lid,_,_) -> lid=id) var_table.lists)

let set_default_val ty =
	match ty with
	Int -> Literal(Int_Literal(0))
	| Double -> Literal(Double_Literal(0.0))
	| Bool -> Literal(Bool_Literal(false))
	| String -> Literal(String_Literal(""))
    | Li(t) -> Literal(Li_Literal([]))
	| _ -> raise(Error("Not a primitive type"))

let get_new_env env func =
	let new_locals = List.fold_left (fun l f -> let (t, id) = (match f with
												Formal(ty, vid) -> (ty, vid)) in
												if check_for_var_existence env.locals id then raise(Error("Variable: " ^ id ^ " already exists in local scope"))
												else (match t with
												| None -> raise(Error("Can't have a variable of type None"))
												| _ -> let new_prims = (id, t, set_default_val t) :: l.prims in {l with prims = new_prims}) ) env.locals func.formals and
	new_functions = func :: env.functions in
	{env with functions = new_functions; locals = new_locals; scope = func.fname; return_type = func.return_type}

let update_prim_table var_table id ty v =
	let does_exist = check_for_var_existence var_table id in
	match does_exist with
	true -> raise(Error("Variable to declare: " ^id^ " already exists"))
	| false ->
		let new_prims = (id, ty, v)::var_table.prims in
			{var_table with prims = new_prims}

let update_list_table var_table id ty v =
	let does_exist = check_for_var_existence var_table id in
	match does_exist with
	true -> raise(Error("Variable to declare already exists"))
	| false ->
		let new_lists = (id, ty, v)::var_table.lists in
	 		{var_table with lists = new_lists}

let rec check_expr env expr =
	match expr with
	| Literal(l) -> get_literal_type l
	| Id(v) -> prerr_string("check_expr: " ^ v ^ " id called\n");
		(try get_type_from_id env.locals v with
		Not_found -> try get_type_from_id env.globals v with
		Not_found -> raise(Error("Id does not appear in program")))
	| Unop(u, e) -> (match u with
		Not -> if check_expr env e = Bool then Bool else raise (Error("Using NOT on a non-boolean expr"))
		| Neg -> if check_expr env e = Double then Double
			else if check_expr env e = Int then Int
			else raise (Error("Using a neg on a non int or float expr")))
	| Binop(e1, op, e2) ->
		let t1 = check_expr env e1 
			and t2 = check_expr env e2 in
		let binop_t = (match op with
			  Add -> check_arithmetic_binary_op t1 t2
			| Sub -> check_arithmetic_binary_op t1 t2
			| Mult -> check_arithmetic_binary_op t1 t2
			| Div -> check_arithmetic_binary_op t1 t2
			| Exp -> check_arithmetic_binary_op t1 t2
			| Equal -> check_equality t1 t2
			| Neq -> check_equality t1 t2
			| Less -> check_logic t1 t2
			| Leq -> check_logic t1 t2
			| Greater -> check_logic t1 t2
			| Geq -> check_logic t1 t2
			| And -> if(t1, t2) = (Bool, Bool) then Bool else raise (Error("Using AND on a non-boolean expr"))
			| Or -> if(t1, t2) = (Bool, Bool) then Bool else raise (Error("Using OR on a non-boolean expr"))
			| Concat -> if (t1, t2) = (String, String) then String else raise (Error("Using Concat on non-string expr"))) in binop_t
	(* built in functions *)
	| Call("print", el) -> prerr_string("Print function is being called\n"); List.iter(fun e -> ignore(check_expr env e) ) el; None
	| Call("rand", el) -> prerr_string("Rand function is being called\n"); List.iter(fun e -> ignore(check_expr env e) ) el; Int
	| Call("str", el) -> prerr_string("Str function is being called\n"); List.iter(fun e -> ignore(check_expr env e) ) el; String
	| Call("evalDouble", el) -> prerr_string("Evaluate function is being called\n"); List.iter(fun e -> ignore(check_expr env e) ) el; Double
	| Call("evalInt", el) -> prerr_string("Evaluate function is being called\n"); List.iter(fun e -> ignore(check_expr env e) ) el; Int
	
	| Call("display_radio", el) -> prerr_string("display radio function is being called\n"); List.iter(fun e -> ignore(check_expr env e) ) el; None
	| Call("display_fillin", el) -> prerr_string("display fillin function is being called\n"); List.iter(fun e -> ignore(check_expr env e) ) el; None
	| Call("get_char_at", el) -> prerr_string("get_char_at function is being called\n"); List.iter(fun e -> ignore(check_expr env e)) el; String
	| Call("length", el) -> prerr_string("length function is being called\n"); List.iter(fun e -> ignore(check_expr env e))el; Int
	| Call("strReplace", el) -> prerr_string("strReplace function is being called\n"); List.iter(fun e -> ignore(check_expr env e)) el; String
	| Call(id, el) -> let func = (try List.find (fun f -> prerr_string("Looking for function: " ^ id ^ " but finding function: " ^f.fname^".\n");
														 f.fname = id) env.functions with
		Not_found -> raise(Error("Function definition not found"))) in
		(try List.iter2 (fun e f -> let ty = (match f with
										Formal(t,_) -> t) in
									let t = check_expr env e in
										if t <> ty then raise(Error("Argument does not match expected argument type"))) el func.formals with
		Invalid_argument s -> raise(Error("Entered the wrong number of arguments into function"))); func.return_type
	| Assign(id, e) -> prerr_string("Assign being called from check_expr\n");
			let tl = (try get_type_from_id env.locals id with
				Not_found -> try get_type_from_id env.globals id with
				Not_found -> raise(Error("Id does not appear in program"))) 
			and tr = check_expr env e in
				prerr_string("tl = " ^ print_type tl ^ " tr = " ^ print_type tr ^ ".\n" );
			if (tl = tr) then
				tl
			else
				raise(Error("Type mismatch in assignment!"))
	
	| Li_list(li) -> check_list_expr li env
	| Li_list_ref(li, ind) -> let t = check_expr env (Id(li)) 
		and i = check_expr env ind in 
		if (i = Int) then 
			t
		else
		raise(Error("Index must be of int type"))
	| Noexpr -> None

and check_list_expr l env  = match l with
| [] -> raise (Error("no empty lists allowed"))
| _ ->
let e = List.hd l in
let t = check_expr env e in 
let s_l = List.map (fun ex -> let t' = check_expr env ex in
				if t = t' then
					t
				else
					raise (Error("Elements of a list literal must all be of the same type"))) l 
in  List.hd s_l

and get_sformal form =
	(match form with
		Formal(ty, s) -> SFormal(ty, s))

and get_sformal_list fl sfl =
	match fl with
	[] -> List.rev sfl
	| head :: tail -> let new_sfl = (get_sformal head) :: sfl in
		get_sformal_list tail new_sfl

and get_slit env ex = match ex with
	Int_Literal(i) -> SInt_Literal(i)
	| Double_Literal(d)-> SDouble_Literal(d)
	| String_Literal(s) -> SString_Literal(s)
	| Bool_Literal(b) -> SBool_Literal(b)
	| _ -> raise(Error("Can't get sliteral of this"))

and get_sexpr env ex = match ex with
	Literal(l) -> (match l with
		Int_Literal(i) -> SLiteral(SInt_Literal(i), Int)
		| Double_Literal(d)-> SLiteral(SDouble_Literal(d), Double)
		| String_Literal(s) -> SLiteral(SString_Literal(s), String)
		| Bool_Literal(b) -> SLiteral(SBool_Literal(b), Bool)
		| _ -> raise(Error("Can't get SLiteral of this")))
	| Id(v) -> SId(v, check_expr env ex)
	| Unop(u, e) -> SUnop(u, get_sexpr env e, check_expr env ex)
	| Binop(e1, op, e2) -> SBinop(get_sexpr env e1, op, get_sexpr env e2, check_expr env ex)
	| Call(str, el) -> SCall(str, List.map (fun e -> get_sexpr env e) el, check_expr env ex)
	| Assign(str, e) -> SAssign(str, get_sexpr env e, check_expr env ex)
	| Li_list(li) -> SLi_list(List.map (fun l -> get_sexpr env l) li, check_expr env ex)
	| Li_list_ref(id, ind) -> SLi_list_ref(id, ind, check_expr env ex)
	| Noexpr -> SNoexpr

and resolve_envs old_env new_env =
	let new_prims = List.map (fun (id, ty, e) -> let v = (try List.find (fun (vid, _, _) -> vid=id) new_env.locals.prims with
		Not_found -> (id, ty, e)) in v) old_env.locals.prims
	and new_lists = List.map (fun (id, ty, el) -> let v = (try List.find (fun (lid, _, _) -> lid=id) new_env.locals.lists with
		Not_found -> (id, ty, el)) in v) old_env.locals.lists in
	{new_env with locals = {prims = new_prims; lists = new_lists}}

and check_stmt env stmt =
	prerr_string("Calling check_stmt\n"); match stmt with
	| Block(stmt_list) ->
		prerr_string("Calling Block from check_stmt\n");
		let new_env = env in
		let (checked_stmts, up_env) = List.fold_left (fun (l, e) s -> let (checked_statment, up_e) = check_stmt e s in
															(checked_statment :: l, up_e)) ([], env) stmt_list in
		let resolved_env = resolve_envs new_env up_env in
		(SBlock(List.rev checked_stmts), resolved_env)
	| Loop(v1, e, v2, s) -> 
		prerr_string("Calling loop from check_stmt\n");
		let (s1, new_env) = check_stmt env s in
		(SLoop((get_sexpr env v1), (get_sexpr env e), (get_sexpr env v2), s1), new_env)
	| Expr(e) -> 
		prerr_string("Calling expression from check_stmt");
		(SExpr(get_sexpr env e), env)
	| Return(e) ->
		prerr_string("Return from check_stmt");
		let t1 = check_expr env e in
			(if not((t1=env.return_type)) then
				raise (Error("Incompatible Return Type using return type: "^ print_type t1 ^" but should be using type "^ print_type env.return_type)));
		let new_env = {env with has_return = true; return_type = t1; return_val = e } in
			(SReturn(get_sexpr env e), new_env)
	| If(e,s1,s2) -> 
		prerr_string("Calling If from check_stmt\n");
		let t1 = check_expr env e in
			(if not(t1=Bool) then
				raise (Error("If statement must be a boolean")));
		let (st1,new_env)= check_stmt env s1 in
		let (st2,new_env2) = check_stmt new_env s2 in
			(SIf((get_sexpr env e), st1, st2),new_env2)
	| Var_decl(decl) -> 
	prerr_string("Calling Var_decl from check_stmt\n");
	let (checked_stmt, up_env) =
		(match decl with
		Var(ty, id) -> prerr_string("Local_Var: Checking " ^ id ^ "\n");
			let new_table = (match ty with
			None -> raise(Error("Can't declare none"))
			| _ -> update_prim_table env.locals id ty (set_default_val ty)) in
			let new_env = {env with locals = new_table} in
			(SVar(ty, id), new_env)
		| Var_Decl_Assign(ty, id, e) -> prerr_string("Var_Decl_Assign: Checking " ^ id ^ " type is "^ Ast.string_of_var_type ty ^ "\n");
			let t_ex = check_expr env e in
			if (t_ex = ty) then
				let new_table = (match ty with
				| Li(t) -> update_list_table env.locals id t []
				| None -> raise(Error("Can't declare none"))
				| _ -> update_prim_table env.locals id ty e) in
				let new_env = {env with locals = new_table} in
				(SVar_Decl_Assign(ty, id, get_sexpr env e), new_env)
			else
				raise(Error("Type mismatch in local variable assignment")))
		 in (SVar_Decl(checked_stmt), up_env)

and get_checked_statements env stmts checked_statments =
	match stmts with
	| stmt :: tail ->
		let (checked_statement, new_env) = check_stmt env stmt in
		get_checked_statements new_env tail (checked_statement::checked_statments)
	| [] -> (List.rev checked_statments, env)

let check_function env func =
	prerr_string("Starting to check function: " ^ func.fname ^ ".\n");
	let (sfstatements, up_env) = get_checked_statements (get_new_env env func) func.body [] in
	({sfname = func.fname; sformals = get_sformal_list func.formals []; sbody = sfstatements; sreturn_type = func.return_type}, {up_env with locals = env.locals})

let rec check_functions env funcs checked_funcs =
	let checked_functions =
		(match funcs with
		func :: tail ->
			let (checked_func, up_env) = check_function env func in
			check_functions up_env tail (checked_func :: checked_funcs)
		| [] -> checked_funcs) in
	checked_functions

let check_global env var =
	let (checked_global, up_env) =
		(match var with
		Var(ty, id) -> prerr_string("Global_Var: Checking " ^ id ^ "\n");
			let new_table = (match ty with
			| None -> raise(Error("Can't declare none"))
			| _ -> update_prim_table env.globals id ty (set_default_val ty)) in
			let new_env = {env with globals = new_table} in
			(SVar(ty, id), new_env)
		| Var_Decl_Assign(ty, id, e) -> prerr_string("Var_Decl_Assign: Checking " ^ id ^ " and type is " ^ Ast.string_of_var_type ty ^ "\n");
			let t_ex = check_expr env e in
			if (t_ex = ty) then
				let new_table = (match ty with
				| Li(t) -> update_list_table env.globals id t []
				| None -> raise(Error("Can't declare none"))
				| _ -> update_prim_table env.globals id ty e) in
				let new_env = {env with globals = new_table} in
				(SVar_Decl_Assign(ty, id, get_sexpr env e), new_env)
			else
				raise(Error("Type mismatch in global variable assignment. Type of t_ex is " ^ Ast.string_of_var_type t_ex )))
		 in
	(checked_global, up_env)

let rec check_globals_and_update_env env vars checked_vars =
	let (checked_globals, new_env) =
		(match vars with
		| var :: tail ->
			let (checked_global, up_env) = check_global env var in
			check_globals_and_update_env up_env tail (checked_global::checked_vars)
		| [] -> (checked_vars, env))
	in (checked_globals, new_env)

let run_program program =
	let (vars, funcs) = program in
	let env = beginning_environment in
	let (checked_globals, new_env) = check_globals_and_update_env env (List.rev vars) [] in
    let checked_functions = check_functions new_env (List.rev funcs) [] in
	SProg(checked_globals, checked_functions)
